import os
import sys

# If this is executed directory, instead of being imported as a class, 
# this file will be inside the './reader/' directory.
sys.path.append('..') 

# from IPython.display import display, HTML

# import logging
from gvec_to_python.util.logger import logger
# logger = logging.getLogger(__name__)

from gvec_to_python.reader.gvec_reader import GVEC_Reader
from gvec_to_python import GVEC

from enum import Enum, unique

@unique
class SampleEnum(Enum):
    """Enum for default sample equilibria."""
    ELLIP_STELL = 1
    ELLIP_STELL_UPDATED = 2
    CIRC_TOK = 3
    ELLIP_STELL_V2_E1D6_M6N6 = 4
    ELLIP_STELL_V2_E4D6_M6N6 = 5
    ELLIP_STELL_V2_E40D5_M6N6 = 6



class SampleEquilibrium():
    """Load a sample GVEC equilibrium within the repository."""

    SAMPLE_DICT = {
        SampleEnum.ELLIP_STELL : {
            'id'      : SampleEnum.ELLIP_STELL,
            'filepath': 'GVEC/testcases/ellipstell',
            'filename': 'GVEC_ellipStell_State_0000_00010000',
            'Nels'    : [21, 30, 78], # num_s=21, num_u=30, num_v=78
            'spl_coef': False,
        },
        SampleEnum.ELLIP_STELL_UPDATED : {
            'id'      : SampleEnum.ELLIP_STELL_UPDATED,
            'filepath': 'GVEC/testcases/ellipstell',
            'filename': 'GVEC_ellipStell_profile_update_State_0000_00010000',
            'Nels'    : [21, 30, 78],
            'spl_coef': True,
        },
        SampleEnum.CIRC_TOK : {
            'id'      : SampleEnum.CIRC_TOK,
            'filepath': 'GVEC/testcases/circ_tok',
            'filename': 'CIRC_TOK_State_0000_00010000',
            'Nels'    : [52, 30, 78],
            'spl_coef': False,
        },
        SampleEnum.ELLIP_STELL_V2_E1D6_M6N6 : {
            'id'      : SampleEnum.ELLIP_STELL_V2_E1D6_M6N6,
            'filepath': 'GVEC/testcases/ellipstell_v2/newBC_E1D6_M6N6',
            'filename': 'GVEC_ELLIPSTELL_V2_State_0000_00200000',
            'Nels'    : [2, 30, 78],
            'spl_coef': False,
        },
        SampleEnum.ELLIP_STELL_V2_E4D6_M6N6 : {
            'id'      : SampleEnum.ELLIP_STELL_V2_E4D6_M6N6,
            'filepath': 'GVEC/testcases/ellipstell_v2/newBC_E4D6_M6N6',
            'filename': 'GVEC_ELLIPSTELL_V2_State_0000_00200000',
            'Nels'    : [5, 30, 78],
            'spl_coef': False,
        },
        SampleEnum.ELLIP_STELL_V2_E40D5_M6N6 : {
            'id'      : SampleEnum.ELLIP_STELL_V2_E40D5_M6N6,
            'filepath': 'GVEC/testcases/ellipstell_v2/oldBC_E40D5M6N6',
            'filename': 'GVEC_ELLIPSTELL_V2_State_0000_00200000',
            'Nels'    : [21, 30, 78],
            'spl_coef': False,
        },
    }



    @classmethod
    def convert_all_to_json(cls) -> None:
        """Ensure that all .dat GVEC samples registed in `SAMPLE_DICT` are converted to .json."""

        basedir = os.path.dirname(os.path.realpath(__file__))

        for sample_id, sample in cls.SAMPLE_DICT.items():
            print(f'Processing {sample_id.name}...')
            filepath = sample['filepath']
            filepath = os.path.join(basedir, '../../..', filepath)
            if not os.path.isfile(os.path.join(filepath, sample['filename'] + '.json')):
                print(f'Sample equilibrium {sample_id.name} has no associated .json file. Generating...')
                reader = GVEC_Reader(filepath, sample['filename'] + '.dat', filepath, sample['filename'] + '.json', with_spl_coef=sample['spl_coef'])
                print(f'JSON file generated for sample equilibrium {sample_id.name}.')



    def __init__(self, sample_enum:SampleEnum, user_path:str=None) -> dict:
        """Load a sample GVEC equilibrium within the repository.

        Parameters
        ----------
        sample_enum : SampleEnum
            Which sample GVEC equilibrium to load.
        user_path : str
            If the path to the samples folder `GVEC` (exclusive) is found elsewhere.
        """

        basedir = os.path.dirname(os.path.realpath(__file__))

        sample = self.SAMPLE_DICT[sample_enum]
        filepath = sample['filepath']
        if user_path is None:
            filepath = os.path.join(basedir, '../../..', filepath)
        else:
            filepath = os.path.join(user_path, filepath)
        if not os.path.isfile(os.path.join(filepath, sample['filename'] + '.json')):
            print(f'Sample equilibrium {sample_enum.name} has no associated .json file. Generating...')
            reader = GVEC_Reader(filepath, sample['filename'] + '.dat', filepath, sample['filename'] + '.json', with_spl_coef=sample['spl_coef'])
            print(f'JSON file generated for sample equilibrium {sample_enum.name}.')

        self._sample = sample
        self._gvec = GVEC(filepath, sample['filename'] + '.json')



    def get_gvec_object(self) -> GVEC:
        """Return an instance of a `GVEC` object of the current sample MHD equilibrium."""
        return self._gvec



    def get_sample_metadata(self) -> dict:
        """Return basic metadata of the current sample MHD equilibrium."""
        return self._sample



    @classmethod
    def get_user_gvec_object(cls, filepath:str, filename:str, with_spl_coef:bool=False) -> GVEC:
        """Return an instance of a `GVEC` object of a user-provided MHD equilibrium."""

        if filename[-4:] == '.dat':
            filename = filename[:-4]
        if not os.path.isfile(os.path.join(filepath, filename + '.json')):
            print(f'Custom equilibrium has no associated .json file. Generating...')
            reader = GVEC_Reader(filepath, filename + '.dat', filepath, filename + '.json', with_spl_coef=with_spl_coef)
            print(f'JSON file generated for custom equilibrium.')
        return GVEC(filepath, filename + '.json')


