# PyTest will consider all `.py` filenames that begin with `test_` (or end with `_test`) as test files.
# Then it will run all functions with prefix `test_`, which may be inside classes with prefix `Test`, and look for `assert` statements.
# Therefore it is important to put all imports inside the function.
# Also because these test functions are located inside the `test` directory, we need to update the import location with `../src`.
# But this is not a good practice. We should install the package in development mode, and run PyTest against that instead.



def test_greville_and_element_boundaries():
    """
    Test reconstructed Greville points and element boundary with GVEC output.
    """

    # ============================================================
    # Imports.
    # ============================================================

    import os
    import sys
    basedir = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, os.path.join(basedir, '..', 'src')) # Because we are inside './test/' directory.
    if sys.version_info[0] < 3:
        from StringIO import StringIO
    else:
        from io import StringIO

    # import logging
    from gvec_to_python.util.logger import logger
    # logger = logging.getLogger(__name__)

    import json
    import numpy as np

    from gvec_to_python.base.base import Base
    from gvec_to_python.base.make_base import make_base
    from gvec_to_python.reader.read_json import read_json
    from gvec_to_python.hmap.suv_to_xyz import suv_to_xyz as MapFull
    from gvec_to_python.hylife.utilities_FEEC import bsplines as bsp
    from gvec_to_python.hylife.utilities_FEEC.spline_space import spline_space_1d

    logger.info(' ')
    logger.info('='*80)
    logger.info('Running `test_greville_and_element_boundaries()`.')
    logger.info('Test reconstructed Greville points and element boundary with GVEC output.')
    logger.info('='*80)
    logger.info(' ')



    # ============================================================
    # Read GVEC JSON output file.
    # ============================================================

    filepath = 'GVEC/testcases/ellipstell/'
    filepath =  os.path.join(basedir, '..', filepath) # Because we are inside './test/' directory.
    filename = 'GVEC_ellipStell_profile_update_State_0000_00010000.json'
    data = read_json(filepath, filename)



    # ============================================================
    # Init GVEC basis.
    # ============================================================

    X1_coef = np.array(data['X1']['coef'])
    X2_coef = np.array(data['X2']['coef'])
    LA_coef = np.array(data['LA']['coef'])
    X1_base = make_base(data, 'X1')
    X2_base = make_base(data, 'X2')
    LA_base = make_base(data, 'LA')



    # ============================================================
    # Init hmap class.
    # ============================================================

    mapfull = MapFull(X1_base, X1_coef, X2_base, X2_coef, LA_base, LA_coef)



    # ============================================================
    # Access profiles from `data`.
    # ============================================================

    spos_grev = np.array(data['profiles']['greville']['spos'])
    assert len(spos_grev) == data['profiles']['greville']['nPoints'],   'Spline profile should have the same length as `nPoints`.'



    # ============================================================
    # Compare reconstructed Grevielle points.
    # ============================================================

    spos_grev     = np.array(spos_grev)
    percent_error = np.zeros_like(spos_grev)
    greville_N    = np.array(X1_base.sbase.greville_N)  # Greville points.
    assert len(greville_N) == len(spos_grev), 'Reconstructed Greville points should have the same size as GVEC ones.'

    logger.info(' ')
    logger.info('Test implementation of Greville points:')
    logger.info('| {: <19} | {: <19} | {: <14} |'.format('   GVEC Grevielle', 'Reconstructed Grev.', '    % Diff'))

    for i, (gvec_grev, reconstr_grev) in enumerate(zip(spos_grev, greville_N)):
        percent_diff     = (reconstr_grev - gvec_grev) / gvec_grev * 100
        percent_error[i] = percent_diff
        logger.info('| {:+.16f} | {:+.16f} | {: 13.8f}% |'.format(
            gvec_grev, reconstr_grev, percent_diff))

    logger.info('| {: <19} | {: <19} | {: <14} |'.format('   GVEC Grevielle', 'Reconstructed Grev.', '    % Diff'))

    logger.info('Maximum error of Greville points except the first point: {: 21.16f}%.'.format(np.nanmax(percent_error[1:])))

    assert np.allclose(spos_grev, greville_N), '`spos` should be equivalent to reconstructed Grevielle points.'

    logger.info(' ')



    # ============================================================
    # Compare implementations of element boundaries.
    # ============================================================

    # Class constructor signature CHANGED in hylife's `devel_standard` branch!
    # The first parameter is now `Nel`, instead of the knot vector!
    # Original code: `self.hylife_sbase = spline_space_1d(self.knots_N, p, bc)`
    # => Remove dependence on STRUPHY code base.

    # Params for s_base.
    el_b = data["grid"]["sGrid"]                    # Element boundaries from GVEC data.
    el_n       = X1_base.sbase.hylife_sbase.el_b    # Current implementation of element boundaries.
    # el_o = bsp.breakpoints(knots_N, p)            # Old STRUPHY implementation
    # el_n = np.linspace(0., 1., Nel + 1)           # New STRUPHY implementation
    # logger.info('Element boundaries as given by GVEC:\n', el_b)
    # logger.info('Element boundaries as given by old spline space:\n', el_o)
    # logger.info('Element boundaries as given by new spline space:\n', el_n)
    percent_error = np.zeros_like(el_b)

    logger.info(' ')
    logger.info('Test implementation of element boundaries:')
    logger.info('| {: <19} | {: <19} | {: <14} |'.format('    GVEC `el_b`', '   Current `el_b`', '    % Diff'))

    for i, (gi, ni) in enumerate(zip(el_b, el_n)):
        percent_diff     = (ni - gi) / gi * 100
        percent_error[i] = percent_diff
        logger.info('| {:+.16f} | {:+.16f} | {: 13.8f}% |'.format(
            gi, ni, percent_diff))

    logger.info('| {: <19} | {: <19} | {: <14} |'.format('    GVEC `el_b`', '   Current `el_b`', '    % Diff'))

    logger.info('Maximum error of element boundary implementations: {: 21.16f}%.'.format(np.nanmax(percent_error)))

    assert np.allclose(el_b, el_n), 'Element boundary is different from reference implementation in GVEC.'

    logger.info(' ')



if __name__ == "__main__":
    test_greville_and_element_boundaries()
