# PyTest will consider all `.py` filenames that begin with `test_` (or end with `_test`) as test files.
# Then it will run all functions with prefix `test_`, which may be inside classes with prefix `Test`, and look for `assert` statements.
# Therefore it is important to put all imports inside the function.
# Also because these test functions are located inside the `test` directory, we need to update the import location with `../src`.
# But this is not a good practice. We should install the package in development mode, and run PyTest against that instead.



def test_implementations_of_dphi():
    """
    Test different implementations of dphi/ds.

    - Compute it analytically
    - Take derivative of the spline
    - Read spline coefficients of dphi/ds profile directly
    - Solve for coefficients from Greville points
    """

    # ============================================================
    # Imports.
    # ============================================================

    import os
    import sys
    basedir = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, os.path.join(basedir, '..', 'src')) # Because we are inside './test/' directory.
    if sys.version_info[0] < 3:
        from StringIO import StringIO
    else:
        from io import StringIO

    # import logging
    from gvec_to_python.util.logger import logger
    # logger = logging.getLogger(__name__)

    import json
    import numpy as np
    import scipy.sparse        as spa
    import scipy.sparse.linalg as spla
    from scipy.sparse.linalg import splu

    import gvec_to_python.hylife.utilities_FEEC.bsplines    as bsp

    from gvec_to_python.base.base import Base
    from gvec_to_python.base.make_base import make_base
    from gvec_to_python.reader.read_json import read_json
    from gvec_to_python.hmap.suv_to_xyz import suv_to_xyz as MapFull
    # from gvec_to_python.hylife.geometry.domain_3d import spline_interpolation_nd

    logger.info(' ')
    logger.info('='*80)
    logger.info('Running `test_implementations_of_dphi()`.')
    logger.info('Compare different implementations of dphi/ds:')
    logger.info('- Compute it analytically')
    logger.info('- Take derivative of the spline')
    logger.info('- Read spline coefficients of dphi/ds profile directly')
    logger.info('- Solve for coefficients from Greville points')
    logger.info('='*80)
    logger.info(' ')



    # ============================================================
    # Read GVEC JSON output file.
    # ============================================================

    filepath = 'GVEC/testcases/ellipstell/'
    filepath =  os.path.join(basedir, '..', filepath) # Because we are inside './test/' directory.
    filename = 'GVEC_ellipStell_profile_update_State_0000_00010000.json'
    data = read_json(filepath, filename)



    # ============================================================
    # Init GVEC basis.
    # ============================================================

    X1_coef = np.array(data['X1']['coef'])
    X2_coef = np.array(data['X2']['coef'])
    LA_coef = np.array(data['LA']['coef'])
    X1_base = make_base(data, 'X1')
    X2_base = make_base(data, 'X2')
    LA_base = make_base(data, 'LA')



    # ============================================================
    # Init hmap class.
    # ============================================================

    mapfull = MapFull(X1_base, X1_coef, X2_base, X2_coef, LA_base, LA_coef)



    # ============================================================
    # Access profiles from `data`.
    # ============================================================

    phi_grev  = np.array(data['profiles']['greville']['phi'])
    chi_grev  = np.array(data['profiles']['greville']['chi'])
    dphi_grev = np.array(data['profiles']['greville']['dphi'])
    dchi_grev = np.array(data['profiles']['greville']['dchi'])
    iota_grev = np.array(data['profiles']['greville']['iota'])
    pres_grev = np.array(data['profiles']['greville']['pres'])
    spos_grev = np.array(data['profiles']['greville']['spos'])

    assert len(phi_grev)  == data['profiles']['greville']['nPoints'],      'Phi greville profile should have the same length as `nPoints`.'
    assert len(chi_grev)  == data['profiles']['greville']['nPoints'],      'Chi greville profile should have the same length as `nPoints`.'
    assert len(dphi_grev) == data['profiles']['greville']['nPoints'],     'dPhi greville profile should have the same length as `nPoints`.'
    assert len(dchi_grev) == data['profiles']['greville']['nPoints'],     'dChi greville profile should have the same length as `nPoints`.'
    assert len(iota_grev) == data['profiles']['greville']['nPoints'],     'Iota greville profile should have the same length as `nPoints`.'
    assert len(pres_grev) == data['profiles']['greville']['nPoints'], 'Pressure greville profile should have the same length as `nPoints`.'
    assert len(spos_grev) == data['profiles']['greville']['nPoints'],   'Spline greville profile should have the same length as `nPoints`.'

    phi_coef  = np.array(data['profiles']['coef']['phi'])
    chi_coef  = np.array(data['profiles']['coef']['chi'])
    dphi_coef = np.array(data['profiles']['coef']['dphi'])
    dchi_coef = np.array(data['profiles']['coef']['dchi'])
    iota_coef = np.array(data['profiles']['coef']['iota'])
    pres_coef = np.array(data['profiles']['coef']['pres'])

    assert len(phi_coef)  == data['profiles']['coef']['nPoints'],      'Phi profile coefficients should have the same length as `nPoints`.'
    assert len(chi_coef)  == data['profiles']['coef']['nPoints'],      'Chi profile coefficients should have the same length as `nPoints`.'
    assert len(dphi_coef) == data['profiles']['coef']['nPoints'],     'dPhi profile coefficients should have the same length as `nPoints`.'
    assert len(dchi_coef) == data['profiles']['coef']['nPoints'],     'dChi profile coefficients should have the same length as `nPoints`.'
    assert len(iota_coef) == data['profiles']['coef']['nPoints'],     'Iota profile coefficients should have the same length as `nPoints`.'
    assert len(pres_coef) == data['profiles']['coef']['nPoints'], 'Pressure profile coefficients should have the same length as `nPoints`.'



    # ============================================================
    # Solve for spline coefficients from Greville points.
    # ============================================================

    # Build collocation matrix from scratch.
    # I_mat = bsp.collocation_matrix(X1_base.sbase.knots_N, X1_base.sbase.p, spos_grev               , periodic=False) # Use Greville points from GVEC data.
    I_mat = bsp.collocation_matrix(X1_base.sbase.knots_N, X1_base.sbase.p, X1_base.sbase.greville_N, periodic=False) # Use Greville points reconstructed from GVEC data.
    I_LU  = splu(spa.csc_matrix(I_mat))

    # Use collocation matrix from `projectors_global_1d`.
    # if not hasattr(X1_base.sbase.hylife_sbase, 'projectors'):
    #     X1_base.sbase.hylife_sbase.set_projectors()
    # colloc = X1_base.sbase.hylife_sbase.projectors.N.toarray()
    # assert np.allclose(colloc, I_mat), 'Collocation matrices generated from two different methods should be identical.'

    # Solve directly.
    phi_coef_from_grev = I_LU.solve(phi_grev)
    dphi_coef_from_grev = I_LU.solve(dphi_grev)

    # Solve using helper function in `domain_3d.py`.
    # Doesn't work because it doesn't use element boundaries to reconstruct the knot vector.
    # phi_coef_from_grev,  T = spline_interpolation_nd([degree], [spos_grev],  phi_grev)
    # dphi_coef_from_grev, T = spline_interpolation_nd([degree], [spos_grev], dphi_grev)



    logger.info(f'phi_coef_from_grev.shape {phi_coef_from_grev.shape}')
    logger.info('{: <29} | {: <29} | {: <19}'.format('          Coefficient', '     Coef from Greville', '       % Diff'))
    phi_cfg_percent_error  = (phi_coef_from_grev - phi_coef) / phi_coef * 100
    for c, cfg, g in zip(phi_coef, phi_coef_from_grev, phi_cfg_percent_error):
        logger.info(f'{c:+29.24f} | {cfg:+29.24f} | {g:+20.14f}%')
    logger.info('{: <29} | {: <29} | {: <19}'.format('          Coefficient', '     Coef from Greville', '       % Diff'))
    logger.info(f'Maximum error when reconstructing control points from greville points for Phi:')
    logger.info(f'{np.nanmax(np.abs( phi_cfg_percent_error)): 20.14f}% at index {np.nanargmax(np.abs( phi_cfg_percent_error)): 2d}.')
    logger.info(' ')



    logger.info(f'dphi_coef_from_grev.shape {dphi_coef_from_grev.shape}')
    logger.info('{: <29} | {: <29} | {: <19}'.format('          Coefficient', '     Coef from Greville', '       % Diff'))
    dphi_cfg_percent_error  = (dphi_coef_from_grev - dphi_coef) / dphi_coef * 100
    for c, cfg, g in zip(dphi_coef, dphi_coef_from_grev, dphi_cfg_percent_error):
        logger.info(f'{c:+29.24f} | {cfg:+29.24f} | {g:+20.14f}%')
    logger.info('{: <29} | {: <29} | {: <19}'.format('          Coefficient', '     Coef from Greville', '       % Diff'))
    logger.info(f'Maximum error when reconstructing control points from greville points for dPhi:')
    logger.info(f'{np.nanmax(np.abs(dphi_cfg_percent_error)): 20.14f}% at index {np.nanargmax(np.abs(dphi_cfg_percent_error)): 2d}.')
    logger.info(' ')



    # ============================================================
    # Interpolate 1D profiles.
    # ============================================================

    def get_phi_1D_analytic(s, base):
        """0-form toroidal flux evaluated analytically."""
        PHIEDGE = base.eval_profile(1, phi_coef)
        evaled = PHIEDGE * s**2
        return evaled

    def get_phi_1D_coefficient(s, base):
        """0-form toroidal flux evaluated given control points of B-spline of X1's sBase."""
        evaled = base.eval_profile(s, phi_coef)
        return evaled

    def get_phi_1D_greville(s, base):
        """0-form toroidal flux evaluated by solving for control points B-spline of X1's sBase from Greville points."""
        evaled = base.eval_profile(s, phi_coef_from_grev)
        return evaled



    # ============================================================
    # Interpolate 1D profiles' first derivatives.
    # ============================================================

    def get_dphi_1D_analytic(s, base):
        """0-form s-derivative of toroidal flux evaluated analytically."""
        PHIEDGE = base.eval_profile(1, phi_coef)
        evaled = PHIEDGE * 2 * s
        return evaled

    def get_dphi_1D_coefficient(s, base):
        """0-form s-derivative of toroidal flux evaluated given control points of B-spline of X1's sBase."""
        evaled = base.eval_profile(s, dphi_coef)
        return evaled

    def get_dphi_1D_greville(s, base):
        """0-form s-derivative of toroidal flux evaluated  by solving for control points B-spline of X1's sBase from Greville points."""
        evaled = base.eval_profile(s, dphi_coef_from_grev)
        return evaled

    def get_dphi_1D_dspline(s, base):
        """0-form s-derivative of toroidal flux evaluated by taking derivative of the B-spline of X1's sBase given control points of phi-profile."""
        evaled = base.eval_dprofile(s, phi_coef)
        return evaled



    # ============================================================
    # Compare implementations of phi.
    # ============================================================

    s_range = np.linspace(1e-12, 1, 101)
    phi_analytic  = np.zeros_like(s_range)
    phi_ctrlpts   = np.zeros_like(s_range)
    phi_greville  = np.zeros_like(s_range)
    dphi_analytic = np.zeros_like(s_range)
    dphi_ctrlpts  = np.zeros_like(s_range)
    dphi_greville = np.zeros_like(s_range)
    dphi_dspline  = np.zeros_like(s_range)
    phi_ctrl_percent_error  = np.zeros_like(s_range)
    phi_grev_percent_error  = np.zeros_like(s_range)
    dphi_ctrl_percent_error = np.zeros_like(s_range)
    dphi_grev_percent_error = np.zeros_like(s_range)
    dphi_dspl_percent_error = np.zeros_like(s_range)

    for i, s in enumerate(s_range):
        phi_analytic[i]  = get_phi_1D_analytic(    s, X1_base)
        phi_ctrlpts[i]   = get_phi_1D_coefficient( s, X1_base)
        phi_greville[i]  = get_phi_1D_greville(    s, X1_base)
        dphi_analytic[i] = get_dphi_1D_analytic(   s, X1_base)
        dphi_ctrlpts[i]  = get_dphi_1D_coefficient(s, X1_base)
        dphi_greville[i] = get_dphi_1D_greville(   s, X1_base)
        dphi_dspline[i]  = get_dphi_1D_dspline(    s, X1_base)

        percent_diff              = (phi_ctrlpts[i]  - phi_analytic[i]) / phi_analytic[i] * 100
        phi_ctrl_percent_error[i] = percent_diff
        percent_diff              = (phi_greville[i] - phi_analytic[i]) / phi_analytic[i] * 100
        phi_grev_percent_error[i] = percent_diff

        percent_diff               = (dphi_ctrlpts[i]  - dphi_analytic[i]) / dphi_analytic[i] * 100
        dphi_ctrl_percent_error[i] = percent_diff
        percent_diff               = (dphi_greville[i] - dphi_analytic[i]) / dphi_analytic[i] * 100
        dphi_grev_percent_error[i] = percent_diff
        percent_diff               = (dphi_dspline[i]  - dphi_analytic[i]) / dphi_analytic[i] * 100
        dphi_dspl_percent_error[i] = percent_diff



    logger.info(' ')
    logger.info('Test implementation of Phi:')
    logger.info('| {: <4} | {: <11} | {: <11} | {: <20} | {: <11} | {: <20} |'.format(' s', 'Phi analyti', ' Phi coef', '       % Diff', ' Phi grev', '       % Diff'))

    for i, s in enumerate(s_range):
        logger.info('| {:.2f} | {:+.8f} | {:+.8f} | {: 19.14f}% | {:+.8f} | {: 19.14f}% |'.format(
            s, phi_analytic[i], phi_ctrlpts[i], phi_ctrl_percent_error[i], phi_greville[i], phi_grev_percent_error[i]))

    logger.info('| {: <4} | {: <11} | {: <11} | {: <20} | {: <11} | {: <20} |'.format(' s', 'Phi analyti', ' Phi coef', '       % Diff', ' Phi grev', '       % Diff'))

    logger.info('Maximum error of Phi against analytical expression:')
    logger.info('From control point : {: 29.24f}% at index {: 3d}.'.format(np.nanmax(np.abs(phi_ctrl_percent_error)), np.nanargmax(np.abs(phi_ctrl_percent_error))))
    logger.info('From greville point: {: 29.24f}% at index {: 3d}.'.format(np.nanmax(np.abs(phi_grev_percent_error)), np.nanargmax(np.abs(phi_grev_percent_error))))

    assert np.allclose(phi_ctrl_percent_error, 0, atol=0.01), 'Phi\'s error should be close to zero, because polynomials are represented exactly by B-splines.'
    assert np.allclose(phi_grev_percent_error, 0, atol=0.1 ), 'Phi\'s error should be close to zero, because polynomials are represented exactly by B-splines.'
    #                                                  ^^^^ Note the extra tolerance!



    logger.info(' ')
    logger.info('Test implementation of dPhi:')
    logger.info('| {: <4} | {: <11} | {: <11} | {: <20} | {: <11} | {: <20} | {: <11} | {: <20} |'.format(' s', 'dPhi analyt', ' dPhi coef', '       % Diff', ' dPhi grev', '       % Diff', ' dPhi dspl', '       % Diff'))

    for i, s in enumerate(s_range):
        logger.info('| {:.2f} | {:+.8f} | {:+.8f} | {: 19.14f}% | {:+.8f} | {: 19.14f}% | {:+.8f} | {: 19.14f}% |'.format(
            s, dphi_analytic[i], dphi_ctrlpts[i], dphi_ctrl_percent_error[i], dphi_greville[i], dphi_grev_percent_error[i], dphi_dspline[i], dphi_dspl_percent_error[i]))

    logger.info('| {: <4} | {: <11} | {: <11} | {: <20} | {: <11} | {: <20} | {: <11} | {: <20} |'.format(' s', 'dPhi analyt', ' dPhi coef', '       % Diff', ' dPhi grev', '       % Diff', ' dPhi dspl', '       % Diff'))

    logger.info('Maximum error of dPhi against analytical expression:')
    logger.info('From control point    : {: 29.24f}% at index {: 3d}.'.format(np.nanmax(np.abs(dphi_ctrl_percent_error)), np.nanargmax(np.abs(dphi_ctrl_percent_error))))
    logger.info('From greville point   : {: 29.24f}% at index {: 3d}.'.format(np.nanmax(np.abs(dphi_grev_percent_error)), np.nanargmax(np.abs(dphi_grev_percent_error))))
    logger.info('From spline derivative: {: 29.24f}% at index {: 3d}.'.format(np.nanmax(np.abs(dphi_dspl_percent_error)), np.nanargmax(np.abs(dphi_dspl_percent_error))))

    assert np.allclose(dphi_ctrl_percent_error, 0, atol=0.01), 'dPhi\'s error should be close to zero, because polynomials are represented exactly by B-splines.'
    assert np.allclose(dphi_grev_percent_error, 0, atol=0.01), 'dPhi\'s error should be close to zero, because polynomials are represented exactly by B-splines.'
    assert np.allclose(dphi_dspl_percent_error, 0, atol=0.01), 'dPhi\'s error should be close to zero, because polynomials are represented exactly by B-splines.'

    logger.info(' ')



if __name__ == "__main__":
    test_implementations_of_dphi()
