# PyTest will consider all `.py` filenames that begin with `test_` (or end with `_test`) as test files.
# Then it will run all functions with prefix `test_`, which may be inside classes with prefix `Test`, and look for `assert` statements.
# Therefore it is important to put all imports inside the function.
# Also because these test functions are located inside the `test` directory, we need to update the import location with `../src`.
# But this is not a good practice. We should install the package in development mode, and run PyTest against that instead.



def test_iota_eq_dchi_dphi():
    """
    Test if Iota profile is equivalent to (dChi/ds) / (dPhi/ds).
    """

    # ============================================================
    # Imports.
    # ============================================================

    import os
    import sys
    basedir = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, os.path.join(basedir, '..', 'src')) # Because we are inside './test/' directory.
    if sys.version_info[0] < 3:
        from StringIO import StringIO
    else:
        from io import StringIO

    # import logging
    from gvec_to_python.util.logger import logger
    # logger = logging.getLogger(__name__)

    import json
    import numpy as np

    from gvec_to_python.base.base import Base
    from gvec_to_python.base.make_base import make_base
    from gvec_to_python.reader.read_json import read_json
    from gvec_to_python.hmap.suv_to_xyz import suv_to_xyz as MapFull

    logger.info(' ')
    logger.info('='*80)
    logger.info('Running `test_iota_eq_dchi_dphi()`.')
    logger.info('Test if Iota profile is equivalent to (dChi/ds) / (dPhi/ds).')
    logger.info('='*80)
    logger.info(' ')



    # ============================================================
    # Read GVEC JSON output file.
    # ============================================================

    filepath = 'GVEC/testcases/ellipstell/'
    filepath =  os.path.join(basedir, '..', filepath) # Because we are inside './test/' directory.
    filename = 'GVEC_ellipStell_profile_update_State_0000_00010000.json'
    data = read_json(filepath, filename)



    # ============================================================
    # Init GVEC basis.
    # ============================================================

    X1_coef = np.array(data['X1']['coef'])
    X2_coef = np.array(data['X2']['coef'])
    LA_coef = np.array(data['LA']['coef'])
    X1_base = make_base(data, 'X1')
    X2_base = make_base(data, 'X2')
    LA_base = make_base(data, 'LA')



    # ============================================================
    # Init hmap class.
    # ============================================================

    mapfull = MapFull(X1_base, X1_coef, X2_base, X2_coef, LA_base, LA_coef)



    # ============================================================
    # Access profiles from `data`.
    # ============================================================

    phi_grev  = np.array(data['profiles']['greville']['phi'])
    chi_grev  = np.array(data['profiles']['greville']['chi'])
    dphi_grev = np.array(data['profiles']['greville']['dphi'])
    dchi_grev = np.array(data['profiles']['greville']['dchi'])
    iota_grev = np.array(data['profiles']['greville']['iota'])
    pres_grev = np.array(data['profiles']['greville']['pres'])
    spos_grev = np.array(data['profiles']['greville']['spos'])

    assert len(phi_grev)  == data['profiles']['greville']['nPoints'],      'Phi greville profile should have the same length as `nPoints`.'
    assert len(chi_grev)  == data['profiles']['greville']['nPoints'],      'Chi greville profile should have the same length as `nPoints`.'
    assert len(dphi_grev) == data['profiles']['greville']['nPoints'],     'dPhi greville profile should have the same length as `nPoints`.'
    assert len(dchi_grev) == data['profiles']['greville']['nPoints'],     'dChi greville profile should have the same length as `nPoints`.'
    assert len(iota_grev) == data['profiles']['greville']['nPoints'],     'Iota greville profile should have the same length as `nPoints`.'
    assert len(pres_grev) == data['profiles']['greville']['nPoints'], 'Pressure greville profile should have the same length as `nPoints`.'
    assert len(spos_grev) == data['profiles']['greville']['nPoints'],   'Spline greville profile should have the same length as `nPoints`.'

    phi_coef  = np.array(data['profiles']['coef']['phi'])
    chi_coef  = np.array(data['profiles']['coef']['chi'])
    dphi_coef = np.array(data['profiles']['coef']['dphi'])
    dchi_coef = np.array(data['profiles']['coef']['dchi'])
    iota_coef = np.array(data['profiles']['coef']['iota'])
    pres_coef = np.array(data['profiles']['coef']['pres'])

    assert len(phi_coef)  == data['profiles']['coef']['nPoints'],      'Phi profile coefficients should have the same length as `nPoints`.'
    assert len(chi_coef)  == data['profiles']['coef']['nPoints'],      'Chi profile coefficients should have the same length as `nPoints`.'
    assert len(dphi_coef) == data['profiles']['coef']['nPoints'],     'dPhi profile coefficients should have the same length as `nPoints`.'
    assert len(dchi_coef) == data['profiles']['coef']['nPoints'],     'dChi profile coefficients should have the same length as `nPoints`.'
    assert len(iota_coef) == data['profiles']['coef']['nPoints'],     'Iota profile coefficients should have the same length as `nPoints`.'
    assert len(pres_coef) == data['profiles']['coef']['nPoints'], 'Pressure profile coefficients should have the same length as `nPoints`.'



    # ============================================================
    # Interpolate 1D profiles.
    # ============================================================
    def get_pres_1D(s, pres_coef, base): # pragma: no cover
        """Return 0-form scalar pressure evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_profile(s, pres_coef)
        return evaled

    def get_phi_1D( s,  phi_coef, base): # pragma: no cover
        """Return 0-form toroidal flux evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_profile(s,  phi_coef)
        return evaled

    def get_chi_1D( s,  chi_coef, base): # pragma: no cover
        """Return 0-form poloidal flux evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_profile(s,  chi_coef)
        return evaled

    def get_iota_1D(s, iota_coef, base):
        """Return 0-form iota profile evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_profile(s, iota_coef)
        return evaled

    def get_spos_1D(s, spos_coef, base): # pragma: no cover
        """Return s-coordinate from Greville points...?"""
        evaled = base.eval_profile(s, spos_coef)
        return evaled



    # ============================================================
    # Interpolate 1D profiles' first derivatives.
    # ============================================================
    def get_dpres_1D(s, pres_coef, base): # pragma: no cover
        """Return 0-form s-derivative of scalar pressure evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_dprofile(s, pres_coef)
        return evaled

    def get_dphi_1D( s,  phi_coef, base):
        """Return 0-form s-derivative of toroidal flux evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_dprofile(s,  phi_coef)
        # evaled = base.eval_profile(s,  dphi_coef) # `dphi_coef` is global here!
        return evaled

    def get_dchi_1D( s,  chi_coef, base):
        """Return 0-form s-derivative of poloidal flux evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_dprofile(s,  chi_coef)
        # evaled = base.eval_profile(s,  dchi_coef) # `dchi_coef` is global here!
        return evaled

    def get_diota_1D(s, iota_coef, base): # pragma: no cover
        """Return s-derivative of 0-form iota profile evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_dprofile(s, iota_coef)
        return evaled

    def get_dspos_1D(s, spos_coef, base): # pragma: no cover
        """Return s-derivative of s-coordinate from Greville points...?"""
        evaled = base.eval_dprofile(s, spos_coef)
        return evaled



    # ============================================================
    # Interpolate 3D 'profiles' (u,v are basically placeholders).
    # ============================================================

    def get_pres_3D(s, u, v, pres_coef, base): # pragma: no cover
        """Return 0-form scalar pressure evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_profile(s, pres_coef)
        evaled = evaled * np.ones_like(u) * np.ones_like(v)
        return evaled

    def get_phi_3D( s, u, v,  phi_coef, base): # pragma: no cover
        """Return 0-form toroidal flux evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_profile(s,  phi_coef)
        evaled = evaled * np.ones_like(u) * np.ones_like(v)
        return evaled

    def get_chi_3D( s, u, v,  chi_coef, base): # pragma: no cover
        """Return 0-form poloidal flux evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_profile(s,  chi_coef)
        evaled = evaled * np.ones_like(u) * np.ones_like(v)
        return evaled

    def get_iota_3D(s, u, v, iota_coef, base): # pragma: no cover
        """Return ?-form iota profile evaluated using B-spline of X1 or X2's sBase."""
        evaled = base.eval_profile(s, iota_coef)
        evaled = evaled * np.ones_like(u) * np.ones_like(v)
        return evaled

    def get_spos_3D(s, u, v, spos_coef, base): # pragma: no cover
        """Return s-coordinate from Greville points...?"""
        evaled = base.eval_profile(s, spos_coef)
        evaled = evaled * np.ones_like(u) * np.ones_like(v)
        return evaled



    # ============================================================
    # Compare iota profile with (dchi/ds) / (dpsi/ds).
    # ============================================================

    s_range = np.linspace(1e-8, 1, 101)
    iota_profile = np.zeros_like(s_range)
    dchidphi_profile = np.zeros_like(s_range)
    percent_error = np.zeros_like(s_range)

    logger.info(' ')
    logger.info('Test implementation of Iota, Chi, Phi profiles:')
    logger.info('| {: <4} | {: <11} | {: <11} | {: <10} | {: <11} | {: <11} |'.format(' s', '   Iota', 'dChi / dPhi', '  % Diff', '   dChi/ds', '   dPhi/ds'))

    for i, s in enumerate(s_range):
        iota = get_iota_1D(s, iota_coef, X1_base)
        dphi = get_dphi_1D(s,  phi_coef, X1_base)
        dchi = get_dchi_1D(s,  chi_coef, X1_base)
        iota_profile[i]     = iota
        dchidphi_profile[i] = dchi / dphi
        percent_diff        = (dchi / dphi - iota) / iota * 100
        percent_error[i]    = percent_diff
        logger.info('| {:.2f} | {:+.8f} | {:+.8f} | {: 9.4f}% | {:+.8f} | {:+.8f} |'.format(
            s, iota, dchi / dphi, percent_diff, dchi, dphi))

    logger.info('| {: <4} | {: <11} | {: <11} | {: <10} | {: <11} | {: <11} |'.format(' s', '   Iota', 'dChi / dPhi', '  % Diff', '   dChi/ds', '   dPhi/ds'))

    logger.info('Maximum error of Iota profile: {: 9.4f}%.'.format(np.max(percent_error)))

    assert np.allclose(iota_profile, dchidphi_profile, rtol=0.01), 'Iota profile should be equivalent to dChi/dPhi, up to a certain precision (1%, typically 0.3652%).'

    logger.info(' ')



if __name__ == "__main__":
    test_iota_eq_dchi_dphi()
