# PyTest will consider all `.py` filenames that begin with `test_` (or end with `_test`) as test files.
# Then it will run all functions with prefix `test_`, which may be inside classes with prefix `Test`, and look for `assert` statements.
# Therefore it is important to put all imports inside the function.
# Also because these test functions are located inside the `test` directory, we need to update the import location with `../src`.
# But this is not a good practice. We should install the package in development mode, and run PyTest against that instead.



def test_nested_maps_and_jacobians_with_meshgrid(num_s=3, num_u=4, num_v=5):
    """
    Test nested `hmap` and their Jacobians with sparse/dense meshgrid.
    """

    # ============================================================
    # Imports.
    # ============================================================

    import os
    import sys
    basedir = os.path.dirname(os.path.realpath(__file__))
    sys.path.insert(0, os.path.join(basedir, '..', 'src')) # Because we are inside './test/' directory.
    if sys.version_info[0] < 3:
        from StringIO import StringIO
    else:
        from io import StringIO

    # import logging
    from gvec_to_python.util.logger import logger
    # logger = logging.getLogger(__name__)

    import warnings
    warnings.filterwarnings("ignore", message="faster on contiguous arrays,")
    # warnings.filterwarnings("ignore", message="NumbaPerformanceWarning: '@' is faster on contiguous arrays,")

    import json
    import numpy as np

    from gvec_to_python.base.base import Base
    from gvec_to_python.base.make_base import make_base
    from gvec_to_python.reader.read_json import read_json
    from gvec_to_python.hmap.suv_to_sthetazeta      import suv_to_sthetazeta      as Map1
    from gvec_to_python.hmap.sthetazeta_to_X1X2zeta import sthetazeta_to_X1X2zeta as Map2
    from gvec_to_python.hmap.X1X2zeta_to_RphiZ      import X1X2zeta_to_RphiZ      as Map3
    from gvec_to_python.hmap.RphiZ_to_xyz           import RphiZ_to_xyz           as Map4
    from gvec_to_python.hmap.suv_to_xyz             import suv_to_xyz             as MapFull

    logger.info(' ')
    logger.info('='*80)
    logger.info('Running `test_nested_maps_and_jacobians_with_meshgrid()`.')
    logger.info('Test nested `hmap` and their Jacobians with sparse/dense meshgrid.')
    logger.info('='*80)
    logger.info(' ')



    # ============================================================
    # Read GVEC JSON output file.
    # ============================================================

    filepath = 'GVEC/testcases/ellipstell/'
    filepath =  os.path.join(basedir, '..', filepath) # Because we are inside './test/' directory.
    filename = 'GVEC_ellipStell_profile_update_State_0000_00010000.json'
    data = read_json(filepath, filename)



    # ============================================================
    # Init GVEC basis.
    # ============================================================

    X1_coef = np.array(data['X1']['coef'])
    X2_coef = np.array(data['X2']['coef'])
    LA_coef = np.array(data['LA']['coef'])
    X1_base = make_base(data, 'X1')
    X2_base = make_base(data, 'X2')
    LA_base = make_base(data, 'LA')



    # ============================================================
    # Init hmap class.
    # ============================================================

    map1    = Map1()
    map2    = Map2(X1_base, X1_coef, X2_base, X2_coef, LA_base, LA_coef)
    map3    = Map3()
    map4    = Map4()
    mapfull = MapFull(X1_base, X1_coef, X2_base, X2_coef, LA_base, LA_coef)
    mapX    = MapFull(X1_base, X1_coef, X2_base, X2_coef, LA_base, LA_coef, component='x')
    mapY    = MapFull(X1_base, X1_coef, X2_base, X2_coef, LA_base, LA_coef, component='y')
    mapZ    = MapFull(X1_base, X1_coef, X2_base, X2_coef, LA_base, LA_coef, component='z')

    def hmap_nested(s, u, v, map2):
        """Transform (s, u, v) coordinates to (x, y, z), via an intermediate (X1, X2, LA).

        Parameters
        ----------
        s : float or meshgrid numpy.ndarray
            Logical coordinate in radial direction.

        u : float or meshgrid numpy.ndarray
            Logical coordinate along poloidal direction. u = theta/(2*PI).

        v : float or meshgrid numpy.ndarray
            Logical coordinate along toroidal direction. v = zeta/(2*PI).

        Returns
        -------
        3-tuple or 3-tuple of meshgrid numpy.ndarray
            The evaluated value for each direction (x, y, z).
        """

        return Map4.mapto(*Map3.mapto(*map2.mapto(*Map1.mapto(s, u, v))))

    def hmap_direct(s, u, v, mapfull):
        """Transform (s, u, v) coordinates to (x, y, z), directly.

        Parameters
        ----------
        s : float or meshgrid numpy.ndarray
            Logical coordinate in radial direction.

        u : float or meshgrid numpy.ndarray
            Logical coordinate along poloidal direction. u = theta/(2*PI).

        v : float or meshgrid numpy.ndarray
            Logical coordinate along toroidal direction. v = zeta/(2*PI).

        Returns
        -------
        3-tuple or 3-tuple of meshgrid numpy.ndarray
            The evaluated value for each direction (x, y, z).
        """

        return mapfull.mapto(s, u, v)

    def hmap_perXYZ(s, u, v, mapX, mapY, mapZ):
        """Transform (s, u, v) coordinates to (x, y, z), directly, but component-wise.

        Parameters
        ----------
        s : float or meshgrid numpy.ndarray
            Logical coordinate in radial direction.

        u : float or meshgrid numpy.ndarray
            Logical coordinate along poloidal direction. u = theta/(2*PI).

        v : float or meshgrid numpy.ndarray
            Logical coordinate along toroidal direction. v = zeta/(2*PI).

        Returns
        -------
        3-tuple or 3-tuple of meshgrid numpy.ndarray
            The evaluated value for each direction (x, y, z).
        """

        return np.array((mapX.mapto(s, u, v), mapY.mapto(s, u, v), mapZ.mapto(s, u, v)))

    def Jacobian_nested(s, u, v, map2):
        """Compute Jacobian at (s, u, v), via an intermediate (X1, X2, LA).

        Parameters
        ----------
        s : float or meshgrid numpy.ndarray
            Logical coordinate in radial direction. Should not be zero.

        u : float or meshgrid numpy.ndarray
            Logical coordinate along poloidal direction. u = theta/(2*PI).

        v : float or meshgrid numpy.ndarray
            Logical coordinate along toroidal direction. v = zeta/(2*PI).

        Returns
        -------
        numpy.ndarray of shape (3, 3) or (..., 3, 3)
            The evaluated value for each direction (x, y, z).
        """

        # Print dimensions to observe weird broadcasting behavior as noted below.
        debug_log = isinstance(s, np.ndarray) and max(s.shape) == s.size

        if debug_log: # pragma: no cover
            logger.info('Jacobian_nested(): Switch on printing debug log below to observe weird broadcasting behavior in intermediate maps, when the input is a sparse meshgrid.')
            logger.info('Jacobian_nested(): This is because, while evaluating X1 and X2 force broadcasting the coordinates to dense meshgrid, zeta remains untouched as sparse.')
            logger.info('Jacobian_nested(): There is no plan to fix this issue, because intermediate maps are not used, and the weird broadcasting does not affect the final assert output.')

        # Turn off `debug_log` by default. Change this to `True` manually to print the array shapes.
        debug_log = debug_log and False

        if debug_log: # pragma: no cover
            print('Map1 inputs: s.shape', s.shape)
            print('Map1 inputs: u.shape', u.shape)
            print('Map1 inputs: v.shape', v.shape)

        Map1_J            = Map1.df(s, u, v)
        Map1_det_direct   = Map1.df_det(s, u, v)
        Map1_det_from_J   = Map1.df_det_from_J(Map1_J)
        Map1_inv_direct   = Map1.df_inv(s, u, v)
        Map1_inv_from_J   = Map1.df_inv_from_J(Map1_J)
        Map1_G_direct     = Map1.G(s, u, v)
        Map1_G_from_J     = Map1.G_from_J(Map1_J)
        Map1_G_inv_direct = Map1.G_inv(s, u, v)
        Map1_G_inv_from_J = Map1.G_inv_from_J(Map1_J)
        Map1_G_inv_from_G = Map1.G_inv_from_G(Map1_G_direct)
        assert np.allclose(Map1_det_direct  , Map1_det_from_J  ), 'Mismatch in Map1 determinant computation.'
        assert np.allclose(Map1_inv_direct  , Map1_inv_from_J  ), 'Mismatch in Map1 Jacobian inverse.'
        assert np.allclose(Map1_G_direct    , Map1_G_from_J    ), 'Mismatch in Map1 Metric tensor G.'
        assert np.allclose(Map1_G_inv_direct, Map1_G_inv_from_J), 'Mismatch in Map1 Metric tensor G inverse.'
        assert np.allclose(Map1_G_inv_direct, Map1_G_inv_from_G), 'Mismatch in Map1 Metric tensor G inverse.'

        if debug_log: # pragma: no cover
            print('Map1_J.shape', Map1_J.shape)

        s, theta, zeta = Map1.mapto(s, u, v)

        if debug_log: # pragma: no cover
            print('Map2 inputs:     s.shape',     s.shape)
            print('Map2 inputs: theta.shape', theta.shape)
            print('Map2 inputs:  zeta.shape',  zeta.shape)

        Map2_J            = map2.df(s, theta, zeta)
        Map2_det_direct   = map2.df_det(s, theta, zeta)
        Map2_det_from_J   = map2.df_det_from_J(Map2_J)
        Map2_inv_direct   = map2.df_inv(s, theta, zeta)
        Map2_inv_from_J   = map2.df_inv_from_J(Map2_J)
        Map2_G_direct     = map2.G(s, theta, zeta)
        Map2_G_from_J     = map2.G_from_J(Map2_J)
        Map2_G_inv_direct = map2.G_inv(s, theta, zeta)
        Map2_G_inv_from_J = map2.G_inv_from_J(Map2_J)
        Map2_G_inv_from_G = map2.G_inv_from_G(Map2_G_direct)
        assert np.allclose(Map2_det_direct  , Map2_det_from_J  ), 'Mismatch in Map2 determinant computation.'
        assert np.allclose(Map2_inv_direct  , Map2_inv_from_J  ), 'Mismatch in Map2 Jacobian inverse.'
        assert np.allclose(Map2_G_direct    , Map2_G_from_J    ), 'Mismatch in Map2 Metric tensor G.'
        assert np.allclose(Map2_G_inv_direct, Map2_G_inv_from_J), 'Mismatch in Map2 Metric tensor G inverse.'
        assert np.allclose(Map2_G_inv_direct, Map2_G_inv_from_G), 'Mismatch in Map2 Metric tensor G inverse.'

        if debug_log: # pragma: no cover
            print('Map2_J.shape', Map2_J.shape)

        X1, X2, zeta = map2.mapto(s, theta, zeta)

        if debug_log: # pragma: no cover
            print('Map3 inputs:   X1.shape',   X1.shape)
            print('Map3 inputs:   X2.shape',   X2.shape)
            print('Map3 inputs: zeta.shape', zeta.shape)

        Map3_J            = Map3.df(X1, X2, zeta)
        Map3_det_direct   = Map3.df_det(X1, X2, zeta)
        Map3_det_from_J   = Map3.df_det_from_J(Map3_J)
        Map3_inv_direct   = Map3.df_inv(X1, X2, zeta)
        Map3_inv_from_J   = Map3.df_inv_from_J(Map3_J)
        Map3_G_direct     = Map3.G(X1, X2, zeta)
        Map3_G_from_J     = Map3.G_from_J(Map3_J)
        Map3_G_inv_direct = Map3.G_inv(X1, X2, zeta)
        Map3_G_inv_from_J = Map3.G_inv_from_J(Map3_J)
        Map3_G_inv_from_G = Map3.G_inv_from_G(Map3_G_direct)
        assert np.allclose(Map3_det_direct  , Map3_det_from_J  ), 'Mismatch in Map3 determinant computation.'
        assert np.allclose(Map3_inv_direct  , Map3_inv_from_J  ), 'Mismatch in Map3 Jacobian inverse.'
        assert np.allclose(Map3_G_direct    , Map3_G_from_J    ), 'Mismatch in Map3 Metric tensor G.'
        assert np.allclose(Map3_G_inv_direct, Map3_G_inv_from_J), 'Mismatch in Map3 Metric tensor G inverse.'
        assert np.allclose(Map3_G_inv_direct, Map3_G_inv_from_G), 'Mismatch in Map3 Metric tensor G inverse.'

        if debug_log: # pragma: no cover
            print('Map3_J.shape', Map3_J.shape)

        R, phi, Z = Map3.mapto(X1, X2, zeta)

        if debug_log: # pragma: no cover
            print('Map4 inputs:   R.shape',   R.shape)
            print('Map4 inputs: phi.shape', phi.shape)
            print('Map4 inputs:   Z.shape',   Z.shape)

        Map4_J            = Map4.df(R, phi, Z)
        Map4_det_direct   = Map4.df_det(R, phi, Z)
        Map4_det_from_J   = Map4.df_det_from_J(Map4_J)
        Map4_inv_direct   = Map4.df_inv(R, phi, Z)
        Map4_inv_from_J   = Map4.df_inv_from_J(Map4_J)
        Map4_G_direct     = Map4.G(R, phi, Z)
        Map4_G_from_J     = Map4.G_from_J(Map4_J)
        Map4_G_inv_direct = Map4.G_inv(R, phi, Z)
        Map4_G_inv_from_J = Map4.G_inv_from_J(Map4_J)
        Map4_G_inv_from_G = Map4.G_inv_from_G(Map4_G_direct)
        assert np.allclose(Map4_det_direct  , Map4_det_from_J  ), 'Mismatch in Map4 determinant computation.'
        assert np.allclose(Map4_inv_direct  , Map4_inv_from_J  ), 'Mismatch in Map4 Jacobian inverse.'
        assert np.allclose(Map4_G_direct    , Map4_G_from_J    ), 'Mismatch in Map4 Metric tensor G.'
        assert np.allclose(Map4_G_inv_direct, Map4_G_inv_from_J), 'Mismatch in Map4 Metric tensor G inverse.'
        assert np.allclose(Map4_G_inv_direct, Map4_G_inv_from_G), 'Mismatch in Map4 Metric tensor G inverse.'

        if debug_log: # pragma: no cover
            print('Map4_J.shape', Map4_J.shape)

        x, y, z = Map4.mapto(R, phi, Z)

        if debug_log: # pragma: no cover
            print('Map4 outputs: x.shape', x.shape)
            print('Map4 outputs: y.shape', y.shape)
            print('Map4 outputs: z.shape', z.shape)

        # Numpy's `matmul()` multiplies the right-most indices, i.e. (..., 3, 3). Use `einsum()` to sum from the left, i.e. (3, 3, ...).
        # Now that all `.df()` functions are returning Jacobians on the right-most indices, we can use `matmul()` again.
        return np.matmul(Map4_J, np.matmul(Map3_J, np.matmul(Map2_J, Map1_J)))
        # return np.einsum('ij...,jk...->ik...', Map4_J, np.einsum('ij...,jk...->ik...', Map3_J, np.einsum('ij...,jk...->ik...', Map2_J, Map1_J)))

    def Jacobian_direct(s, u, v, mapfull):
        """Compute Jacobian at (s, u, v), directly.

        Parameters
        ----------
        s : float or meshgrid numpy.ndarray
            Logical coordinate in radial direction. Should not be zero.

        u : float or meshgrid numpy.ndarray
            Logical coordinate along poloidal direction. u = theta/(2*PI).

        v : float or meshgrid numpy.ndarray
            Logical coordinate along toroidal direction. v = zeta/(2*PI).

        Returns
        -------
        numpy.ndarray of shape (3, 3) or (..., 3, 3)
            The evaluated value for each direction (x, y, z).
        """

        MapFull_J            = mapfull.df(s, u, v)
        MapFull_det_direct   = mapfull.df_det(s, u, v)
        MapFull_det_from_J   = mapfull.df_det_from_J(MapFull_J)
        MapFull_inv_direct   = mapfull.df_inv(s, u, v)
        MapFull_inv_from_J   = mapfull.df_inv_from_J(MapFull_J)
        MapFull_G_direct     = mapfull.G(s, u, v)
        MapFull_G_from_J     = mapfull.G_from_J(MapFull_J)
        MapFull_G_inv_direct = mapfull.G_inv(s, u, v)
        MapFull_G_inv_from_J = mapfull.G_inv_from_J(MapFull_J)
        MapFull_G_inv_from_G = mapfull.G_inv_from_G(MapFull_G_direct)
        assert np.allclose(MapFull_det_direct  , MapFull_det_from_J  ), 'Mismatch in MapFull determinant computation.'
        assert np.allclose(MapFull_inv_direct  , MapFull_inv_from_J  ), 'Mismatch in MapFull Jacobian inverse.'
        assert np.allclose(MapFull_G_direct    , MapFull_G_from_J    ), 'Mismatch in MapFull Metric tensor G.'
        assert np.allclose(MapFull_G_inv_direct, MapFull_G_inv_from_J), 'Mismatch in MapFull Metric tensor G inverse.'
        assert np.allclose(MapFull_G_inv_direct, MapFull_G_inv_from_G), 'Mismatch in MapFull Metric tensor G inverse.'

        return MapFull_J



    # ============================================================
    # Setup testing grid.
    # ============================================================

    s_range = np.linspace(0, 1, num_s+2)[1:] # Skip 0.
    u_range = np.linspace(0, 1, num_u+1)
    v_range = np.linspace(0, 1, num_v+1)
    logger.info('Total number of points: {}'.format(s_range.shape[0] * u_range.shape[0] * v_range.shape[0]))
    logger.info(' ')

    s_sparse, u_sparse, v_sparse = np.meshgrid(s_range, u_range, v_range, indexing='ij', sparse=True)
    s_dense,  u_dense,  v_dense  = np.meshgrid(s_range, u_range, v_range, indexing='ij', sparse=False)
    logger.info('s_range.shape: {}'.format(s_range.shape,))
    logger.info('u_range.shape: {}'.format(u_range.shape,))
    logger.info('v_range.shape: {}'.format(v_range.shape,))
    logger.info(' ')
    logger.info('s_sparse.shape: {}'.format(s_sparse.shape,))
    logger.info('u_sparse.shape: {}'.format(u_sparse.shape,))
    logger.info('v_sparse.shape: {}'.format(v_sparse.shape,))
    logger.info(' ')
    logger.info('s_dense.shape: {}'.format(s_dense.shape,))
    logger.info('u_dense.shape: {}'.format(u_dense.shape,))
    logger.info('v_dense.shape: {}'.format(v_dense.shape,))
    logger.info(' ')



    # ============================================================
    # Test hmap: Scalar inputs.
    # ============================================================

    logger.info('Test hmap: Scalar inputs: Starts.')

    hmap_nested_scalar = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3))
    hmap_direct_scalar = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3))
    hmap_perXYZ_scalar = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3))

    for s_idx, s in enumerate(s_range):
        for u_idx, u in enumerate(u_range):
            for v_idx, v in enumerate(v_range):
                hmap_nested_scalar[s_idx, u_idx, v_idx, :] = hmap_nested(s, u, v, map2)
                hmap_direct_scalar[s_idx, u_idx, v_idx, :] = hmap_direct(s, u, v, mapfull)
                hmap_perXYZ_scalar[s_idx, u_idx, v_idx, :] = hmap_perXYZ(s, u, v, mapX, mapY, mapZ)

    logger.info('Nested map shape: {}'.format(hmap_nested_scalar.shape,))
    logger.info('Direct map shape: {}'.format(hmap_direct_scalar.shape,))
    logger.info('PerXYZ map shape: {}'.format(hmap_perXYZ_scalar.shape,))
    assert np.allclose(hmap_nested_scalar, hmap_direct_scalar), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(hmap_nested_scalar, hmap_perXYZ_scalar), 'Failed to establish equivalence of the two maps.'

    hmap_nested_swapped = np.moveaxis(hmap_nested_scalar, -1, 0)
    hmap_direct_swapped = np.moveaxis(hmap_direct_scalar, -1, 0)
    logger.info('Nested map shape (moveaxis): {}'.format(hmap_nested_swapped.shape,))
    logger.info('Direct map shape (moveaxis): {}'.format(hmap_direct_swapped.shape,))

    logger.info('Test hmap: Scalar inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test hmap: Scalar inputs as 0D arrays.
    # ============================================================

    logger.info('Test hmap: Scalar inputs as 0D arrays: Starts.')

    hmap_nested_scalar0D = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3))
    hmap_direct_scalar0D = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3))
    hmap_perXYZ_scalar0D = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3))

    for s_idx, s in enumerate(s_range):
        for u_idx, u in enumerate(u_range):
            for v_idx, v in enumerate(v_range):
                hmap_nested_scalar0D[s_idx, u_idx, v_idx, :] = hmap_nested(np.array(s), np.array(u), np.array(v), map2)
                hmap_direct_scalar0D[s_idx, u_idx, v_idx, :] = hmap_direct(np.array(s), np.array(u), np.array(v), mapfull)
                hmap_perXYZ_scalar0D[s_idx, u_idx, v_idx, :] = hmap_perXYZ(np.array(s), np.array(u), np.array(v), mapX, mapY, mapZ)

    logger.info('Nested map shape: {}'.format(hmap_nested_scalar0D.shape,))
    logger.info('Direct map shape: {}'.format(hmap_direct_scalar0D.shape,))
    logger.info('PerXYZ map shape: {}'.format(hmap_perXYZ_scalar0D.shape,))
    assert np.allclose(hmap_nested_scalar0D, hmap_direct_scalar0D), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(hmap_nested_scalar0D, hmap_perXYZ_scalar0D), 'Failed to establish equivalence of the two maps.'

    hmap_nested_swapped0D = np.moveaxis(hmap_nested_scalar0D, -1, 0)
    hmap_direct_swapped0D = np.moveaxis(hmap_direct_scalar0D, -1, 0)
    logger.info('Nested map shape (moveaxis): {}'.format(hmap_nested_swapped0D.shape,))
    logger.info('Direct map shape (moveaxis): {}'.format(hmap_direct_swapped0D.shape,))

    logger.info('Test hmap: Scalar inputs as 0D arrays: Completed.')
    logger.info(' ')



    # ============================================================
    # Test hmap: 1D arrays before meshgrid.
    # ============================================================

    logger.info('Test hmap: 1D arrays before meshgrid inputs: Starts.')

    hmap_nested_range = hmap_nested(s_range, u_range, v_range, map2)
    hmap_direct_range = hmap_direct(s_range, u_range, v_range, mapfull)
    hmap_perXYZ_range = hmap_perXYZ(s_range, u_range, v_range, mapX, mapY, mapZ)

    logger.info('Nested map shape: {}'.format(hmap_nested_range.shape,))
    logger.info('Direct map shape: {}'.format(hmap_direct_range.shape,))
    logger.info('PerXYZ map shape: {}'.format(hmap_perXYZ_range.shape,))
    assert np.allclose(hmap_nested_range, hmap_direct_range), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(hmap_nested_range, hmap_perXYZ_range), 'Failed to establish equivalence of the two maps.'

    logger.info('Test hmap: 1D arrays before meshgrid inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test hmap: Sparse meshgrid.
    # ============================================================

    logger.info('Test hmap: Sparse meshgrid inputs: Starts.')

    hmap_nested_sparse = hmap_nested(s_sparse, u_sparse, v_sparse, map2)
    hmap_direct_sparse = hmap_direct(s_sparse, u_sparse, v_sparse, mapfull)
    hmap_perXYZ_sparse = hmap_perXYZ(s_sparse, u_sparse, v_sparse, mapX, mapY, mapZ)

    logger.info('Nested map shape: {}'.format(hmap_nested_sparse.shape,))
    logger.info('Direct map shape: {}'.format(hmap_direct_sparse.shape,))
    logger.info('PerXYZ map shape: {}'.format(hmap_perXYZ_sparse.shape,))
    assert np.allclose(hmap_nested_sparse, hmap_direct_sparse), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(hmap_nested_sparse, hmap_perXYZ_sparse), 'Failed to establish equivalence of the two maps.'

    logger.info('Test hmap: Sparse meshgrid inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test hmap: Dense meshgrid.
    # ============================================================

    logger.info('Test hmap: Dense meshgrid inputs: Starts.')

    hmap_nested_dense = hmap_nested(s_dense, u_dense, v_dense, map2)
    hmap_direct_dense = hmap_direct(s_dense, u_dense, v_dense, mapfull)
    hmap_perXYZ_dense = hmap_perXYZ(s_dense, u_dense, v_dense, mapX, mapY, mapZ)

    logger.info('Nested map shape: {}'.format(hmap_nested_dense.shape,))
    logger.info('Direct map shape: {}'.format(hmap_direct_dense.shape,))
    logger.info('PerXYZ map shape: {}'.format(hmap_perXYZ_dense.shape,))
    assert np.allclose(hmap_nested_dense, hmap_direct_dense), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(hmap_nested_dense, hmap_perXYZ_dense), 'Failed to establish equivalence of the two maps.'

    logger.info('Test hmap: Dense meshgrid inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test hmap: Check results from different techniques match.
    # ============================================================

    logger.info('Test hmap: Compare among inputs: Starts.')

    assert np.allclose(hmap_direct_swapped,   hmap_direct_dense), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(hmap_direct_swapped0D, hmap_direct_dense), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(hmap_direct_range,     hmap_direct_dense), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(hmap_direct_sparse,    hmap_direct_dense), 'Failed to establish equivalence of the two maps.'

    logger.info('Test hmap: Compare among inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test Jacobian: Scalar inputs.
    # ============================================================

    logger.info('Test Jacobian: Scalar inputs: Starts.')

    Jacobian_nested_scalar = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3, 3))
    Jacobian_direct_scalar = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3, 3))

    for s_idx, s in enumerate(s_range):
        for u_idx, u in enumerate(u_range):
            for v_idx, v in enumerate(v_range):
                Jacobian_nested_scalar[s_idx, u_idx, v_idx, :, :] = Jacobian_nested(s, u, v, map2)
                Jacobian_direct_scalar[s_idx, u_idx, v_idx, :, :] = Jacobian_direct(s, u, v, mapfull)

    logger.info('Nested map shape: {}'.format(Jacobian_nested_scalar.shape,))
    logger.info('Direct map shape: {}'.format(Jacobian_direct_scalar.shape,))
    assert np.allclose(Jacobian_nested_scalar, Jacobian_direct_scalar), 'Failed to establish equivalence of the two maps\' Jacobians.'

    logger.info('Test Jacobian: Scalar inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test Jacobian: Scalar inputs as 0D arrays.
    # ============================================================

    logger.info('Test Jacobian: Scalar inputs as 0D arrays: Starts.')

    Jacobian_nested_scalar0D = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3, 3))
    Jacobian_direct_scalar0D = np.zeros((s_range.shape[0], u_range.shape[0], v_range.shape[0], 3, 3))

    for s_idx, s in enumerate(s_range):
        for u_idx, u in enumerate(u_range):
            for v_idx, v in enumerate(v_range):
                Jacobian_nested_scalar0D[s_idx, u_idx, v_idx, :, :] = Jacobian_nested(s, u, v, map2)
                Jacobian_direct_scalar0D[s_idx, u_idx, v_idx, :, :] = Jacobian_direct(s, u, v, mapfull)

    logger.info('Nested map shape: {}'.format(Jacobian_nested_scalar0D.shape,))
    logger.info('Direct map shape: {}'.format(Jacobian_direct_scalar0D.shape,))
    assert np.allclose(Jacobian_nested_scalar0D, Jacobian_direct_scalar0D), 'Failed to establish equivalence of the two maps\' Jacobians.'

    logger.info('Test Jacobian: Scalar inputs as 0D arrays: Completed.')
    logger.info(' ')



    # ============================================================
    # Test Jacobian: 1D arrays before meshgrid.
    # ============================================================

    logger.info('Test Jacobian: 1D arrays before meshgrid inputs: Starts.')

    Jacobian_nested_range = Jacobian_nested(s_range, u_range, v_range, map2)
    Jacobian_direct_range = Jacobian_direct(s_range, u_range, v_range, mapfull)

    logger.info('Nested map shape: {}'.format(Jacobian_nested_range.shape,))
    logger.info('Direct map shape: {}'.format(Jacobian_direct_range.shape,))
    assert np.allclose(Jacobian_nested_range, Jacobian_direct_range), 'Failed to establish equivalence of the two maps.'

    logger.info('Test Jacobian: 1D arrays before meshgrid inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test Jacobian: Sparse meshgrid.
    # ============================================================

    logger.info('Test Jacobian: Sparse meshgrid inputs: Starts.')

    Jacobian_nested_sparse = Jacobian_nested(s_sparse, u_sparse, v_sparse, map2)
    Jacobian_direct_sparse = Jacobian_direct(s_sparse, u_sparse, v_sparse, mapfull)

    logger.info('Nested map shape: {}'.format(Jacobian_nested_sparse.shape,))
    logger.info('Direct map shape: {}'.format(Jacobian_direct_sparse.shape,))
    assert np.allclose(Jacobian_nested_sparse, Jacobian_direct_sparse), 'Failed to establish equivalence of the two maps.'

    logger.info('Test Jacobian: Sparse meshgrid inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test Jacobian: Dense meshgrid.
    # ============================================================

    logger.info('Test Jacobian: Dense meshgrid inputs: Starts.')

    Jacobian_nested_dense = Jacobian_nested(s_dense, u_dense, v_dense, map2)
    Jacobian_direct_dense = Jacobian_direct(s_dense, u_dense, v_dense, mapfull)

    logger.info('Nested map shape: {}'.format(Jacobian_nested_dense.shape,))
    logger.info('Direct map shape: {}'.format(Jacobian_direct_dense.shape,))
    assert np.allclose(Jacobian_nested_dense, Jacobian_direct_dense), 'Failed to establish equivalence of the two maps.'

    logger.info('Test Jacobian: Dense meshgrid inputs: Completed.')
    logger.info(' ')



    # ============================================================
    # Test Jacobian: Check results from different techniques match.
    # ============================================================

    logger.info('Test Jacobian: Compare among inputs: Starts.')

    assert np.allclose(Jacobian_direct_scalar,   Jacobian_direct_dense), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(Jacobian_direct_scalar0D, Jacobian_direct_dense), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(Jacobian_direct_range,    Jacobian_direct_dense), 'Failed to establish equivalence of the two maps.'
    assert np.allclose(Jacobian_direct_sparse,   Jacobian_direct_dense), 'Failed to establish equivalence of the two maps.'

    logger.info('Test Jacobian: Compare among inputs: Completed.')
    logger.info(' ')



if __name__ == "__main__":
    test_nested_maps_and_jacobians_with_meshgrid()
